@@ -366,13 +366,19 @@ static int
 meta_showpage (GnomePrintContext *pc)
 {
 	encode_int (pc, GNOME_META_SHOWPAGE);
+	GNOME_PRINT_META (pc)->pages++;
 	return 0;
 }
 
 static int
 meta_close (GnomePrintContext *pc)
 {
+	GnomePrintMeta *meta = GNOME_PRINT_META (pc);
+
 	encode_int (pc, GNOME_META_CLOSE);
+
+	((GnomeMetaFileHeader *)meta->buffer)->size = meta->buffer_size;
+
 	return 0;
 }
 
@@ -493,7 +499,7 @@ gnome_print_meta_new_from (void *data)
 
 	g_return_val_if_fail (data != NULL, NULL);
 
-	if (strcmp (data, GNOME_METAFILE_SIGNATURE) != 0)
+	if (strcmp ((char *)data, GNOME_METAFILE_SIGNATURE) != 0)
 		return NULL;
 
 	len = mf->size;
@@ -564,22 +570,28 @@ gnome_print_meta_get_copy (GnomePrintMeta *meta, void **buffer, int *buflen)
 	return 1;
 }
 
-static char *
-decode_int (char *data, gint32 *dest)
+static const char *
+decode_int (const char *data, gint32 *dest)
 {
-	*dest = g_ntohl (*(int *) data);
+	gint32 nint;
+	memcpy (&nint, data, sizeof (gint32));
+	*dest = g_ntohl (nint);
 	return data + sizeof (gint32);
 }
 
-static char *
-decode_double (char *data, double *dest)
+static const char *
+decode_double (const char *data, double *dest)
 {
-	*dest = *(double *) data;
+	/* Since the size is known, this should not be as bad as it
+	   looks.  Check the generated assembler before you come up
+	   with non-ANSI hacks.  -- MW.  */
+	/* Should we do something like g_ntohl?  */
+	memcpy (dest, data, sizeof (double));
 	return data + sizeof (double);
 }
 
-static char *
-decode_image (char *data, GnomePrintContext *dest, int bytes_per_pixel)
+static const char *
+decode_image (const char *data, GnomePrintContext *dest, int bytes_per_pixel, gboolean do_output)
 {
 	void *buffer;
 	gint32 height, width;
@@ -589,26 +601,38 @@ decode_image (char *data, GnomePrintContext *dest, int bytes_per_pixel)
 	data = decode_int (data, &width);
 
 	size = height * width * bytes_per_pixel;
-	buffer = g_malloc (size);
-	memcpy (buffer, data, size);
-
-	if (bytes_per_pixel == 1)
-		gnome_print_grayimage (dest, buffer, width, height, width);
-	else
-		gnome_print_rgbimage (dest, buffer, width, height, width * bytes_per_pixel);
-
-	g_free (buffer);
+	if (do_output){
+		buffer = g_malloc (size);
+		memcpy (buffer, data, size);
+		
+		if (bytes_per_pixel == 1)
+			gnome_print_grayimage (dest, buffer, width, height, width);
+		else
+			gnome_print_rgbimage (dest, buffer, width, height, width * bytes_per_pixel);
+		
+		g_free (buffer);
+	}
 	return data + size;
 }
 
 static gboolean
-do_render (GnomePrintContext *dest, char *data, int size)
+do_render (GnomePrintContext *dest, const char *data, int size, int page)
 {
-	char *end = data + size;
+	const char *end = data + size;
 	double matrix [6];
 	double x1, y1, x2, y2, x3, y3;
 	double x, y;
 	double r, g, b;
+	int    current_page = 0;
+	gboolean do_output;
+	
+	/*
+	 * Output if all pages were requested, or the first page was
+	 */
+	if (page == -1 || page == 0)
+		do_output = TRUE;
+	else
+		do_output = FALSE;
 	
 	while (data < end){
 		gint32 opcode, i;
@@ -616,19 +640,22 @@ do_render (GnomePrintContext *dest, char *data, int size)
 		data = decode_int (data, &opcode);
 		switch ((GnomeMetaType) opcode){
 		case GNOME_META_NEWPATH:
-			gnome_print_newpath (dest);
+			if (do_output)
+				gnome_print_newpath (dest);
 			break;
 			
 		case GNOME_META_MOVETO:
 			data = decode_double (data, &x);
 			data = decode_double (data, &y);
-			gnome_print_moveto (dest, x, y);
+			if (do_output)
+				gnome_print_moveto (dest, x, y);
 			break;
 
 		case GNOME_META_LINETO:
 			data = decode_double (data, &x);
 			data = decode_double (data, &y);
-			gnome_print_lineto (dest, x, y);
+			if (do_output)
+				gnome_print_lineto (dest, x, y);
 			break;
 			
 		case GNOME_META_CURVETO:
@@ -638,11 +665,14 @@ do_render (GnomePrintContext *dest, char *data, int size)
 			data = decode_double (data, &y2);
 			data = decode_double (data, &x3);
 			data = decode_double (data, &y3);
-			gnome_print_curveto (dest, x1, y1, x2, y2, x3, y3);
+
+			if (do_output)
+				gnome_print_curveto (dest, x1, y1, x2, y2, x3, y3);
 			break;
 			
 		case GNOME_META_CLOSEPATH:
-			gnome_print_closepath (dest);
+			if (do_output)
+				gnome_print_closepath (dest);
 			break;
 
 		case GNOME_META_SETRGBCOLOR:
@@ -653,11 +683,13 @@ do_render (GnomePrintContext *dest, char *data, int size)
 			break;
 			
 		case GNOME_META_FILL:
-			gnome_print_fill (dest);
+			if (do_output)
+				gnome_print_fill (dest);
 			break;
 			
 		case GNOME_META_EOFILL:
-			gnome_print_eofill (dest);
+			if (do_output)
+				gnome_print_eofill (dest);
 			break;
 			
 		case GNOME_META_SETLINEWIDTH:
@@ -695,11 +727,13 @@ do_render (GnomePrintContext *dest, char *data, int size)
 		}
 		
 		case GNOME_META_STROKEPATH:
-			gnome_print_strokepath (dest);
+			if (do_output)
+				gnome_print_strokepath (dest);
 			break;
 			
 		case GNOME_META_STROKE:
-			gnome_print_stroke (dest);
+			if (do_output)
+				gnome_print_stroke (dest);
 			break;
 			
 		case GNOME_META_SETFONT: {
@@ -720,8 +754,6 @@ do_render (GnomePrintContext *dest, char *data, int size)
 				gnome_print_setfont (dest, font);
 				gtk_object_unref (GTK_OBJECT (font));
 			}
-				
-			g_warning ("GNOME_META_SETFONT: No fallback code defined\n");
 			break;
 		}
 			
@@ -733,7 +765,8 @@ do_render (GnomePrintContext *dest, char *data, int size)
 			memcpy (buffer, data, i);
 			buffer [i] = 0;
 			data += i;
-			gnome_print_show (dest, buffer);
+			if (do_output)
+				gnome_print_show (dest, buffer);
 			g_free (buffer);
 			break;
 		}
@@ -769,15 +802,21 @@ do_render (GnomePrintContext *dest, char *data, int size)
 			break;
 			
 		case GNOME_META_GRAYIMAGE:
-			data = decode_image (data, dest, 1);
+			data = decode_image (data, dest, 1, do_output);
 			break;
 			
 		case GNOME_META_RGBIMAGE:
-			data = decode_image (data, dest, 3);
+			data = decode_image (data, dest, 3, do_output);
 			break;
 			
 		case GNOME_META_SHOWPAGE:
 			gnome_print_showpage (dest);
+			if (page == current_page){
+				return TRUE;
+			}
+			current_page++;
+			if (current_page == page)
+				do_output = TRUE;
 			break;
 
 		case GNOME_META_TEXTLINE:
@@ -810,12 +849,12 @@ gnome_print_meta_render (GnomePrintContext *destination, void *meta_stream)
 	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (destination), FALSE);
 	g_return_val_if_fail (meta_stream != NULL, FALSE);
 
-	if (strcmp (meta_stream, GNOME_METAFILE_SIGNATURE) != 0)
+	if (strcmp ((char *) meta_stream, GNOME_METAFILE_SIGNATURE) != 0)
 		return FALSE;
 
 	data = meta_stream;
 	size = ((GnomeMetaFileHeader *)meta_stream)->size;
-	return do_render (destination, data + sizeof (GnomeMetaFileHeader), size);
+	return do_render (destination, data + sizeof (GnomeMetaFileHeader), size, -1);
 }
 
 /**
@@ -839,3 +878,63 @@ gnome_print_meta_render_from_object (GnomePrintContext *destination, GnomePrintM
 	return gnome_print_meta_render (destination, source->buffer);
 }
 
+/**
+ * gnome_print_meta_render:
+ * @destination: Destination printer context.
+ * @meta_stream: a metadata stream to replay
+ * @page: Page to be rendered.
+ *
+ * Plays the @meta_steam metadata stream into the @destination printer.
+ * Output commands will only take place for page @page.
+ *
+ * Returns TRUE on success.
+ */
+gboolean
+gnome_print_meta_render_page (GnomePrintContext *destination, void *meta_stream, int page)
+{
+	char *data;
+	int size;
+	
+	g_return_val_if_fail (destination != NULL, FALSE);
+	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (destination), FALSE);
+	g_return_val_if_fail (meta_stream != NULL, FALSE);
+
+	if (strcmp ((char *) meta_stream, GNOME_METAFILE_SIGNATURE) != 0)
+		return FALSE;
+
+	data = meta_stream;
+	size = ((GnomeMetaFileHeader *)meta_stream)->size;
+	return do_render (destination, data + sizeof (GnomeMetaFileHeader), size, page);
+}
+
+/**
+ * gnome_print_meta_render_from_object
+ * @destination: Destination printer context.
+ * @source: an existing GnomePrintMeta printer context.
+ * @page: Page to be rendered.
+ *
+ * Plays the contents of the GnomePrintMeta @source metadata stream
+ * into the @destination printer.  Output commands will only take place for page @page.
+ *
+ * Returns TRUE on success.
+ */
+gboolean
+gnome_print_meta_render_from_object_page (GnomePrintContext *destination, GnomePrintMeta *source, int page)
+{
+	g_return_val_if_fail (destination != NULL, FALSE);
+	g_return_val_if_fail (source != NULL, FALSE);
+	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (destination), FALSE);
+	g_return_val_if_fail (GNOME_IS_PRINT_META (source), FALSE);
+
+	return gnome_print_meta_render_page (destination, source->buffer, page);
+}
+
+int
+gnome_print_meta_pages (GnomePrintMeta *meta)
+{
+	g_return_val_if_fail (meta != NULL, 0);
+	g_return_val_if_fail (GNOME_IS_PRINT_META (meta), 0);
+
+	return meta->pages;
+}
+
