@@ -285,11 +285,24 @@ gpp_stroke (GnomePrintContext *pc)
 	GnomePrintPreviewPrivate *priv = pp->priv;
 	GnomeCanvasItem *item;
 	GraphicContext *g = gc (priv);
+	int close;
 	
 	if (g->bpath == NULL)
 		return 0;
-	
-	close_path (g->bpath);
+
+	/*
+	 * Special case: libart does not like stroking of lines
+	 */
+	close = 1;
+	if (g->bpath->n_bpath == 2){
+		if ((g->bpath->bpath [0].code == ART_MOVETO ||
+		     g->bpath->bpath [0].code == ART_MOVETO_OPEN) &&
+		    g->bpath->bpath [1].code == ART_LINETO)
+			close = 0;
+	}
+
+	if (close)
+		close_path (g->bpath);
 	lazy_showpage_check (priv);
 	item = gnome_canvas_item_new (
 		GNOME_CANVAS_GROUP (priv->page),
@@ -645,9 +658,8 @@ gpp_showpage (GnomePrintContext *pc)
 	GnomePrintPreviewPrivate *priv = pp->priv;
 
 	g_array_append_val (priv->pages, priv->page);
-#if 0
-	gnome_canvas_item_hide (GNOME_CANVAS_ITEM (priv->page));
-#endif
+
+	priv->auto_hide_page_on_stroke = priv->page;
 	
 	priv->page = gnome_canvas_item_new (
 		gnome_canvas_root (pp->canvas),
@@ -770,16 +782,27 @@ gpp_finalize (GtkObject *object)
 	GnomePrintPreview *pp = GNOME_PRINT_PREVIEW (object);
 	GnomePrintPreviewPrivate *priv = pp->priv;
 	GSList *l;
+	int i;
 
 	for (l = priv->context_list; l; l = l->next)
 		gc_free (l->data);
 	
-	g_array_free (priv->pages, TRUE);
-
 	if (pp->canvas)
 		gtk_object_unref (GTK_OBJECT (pp->canvas));
 
-	g_hash_table_foreach_remove (priv->font_hash, kill_font, NULL);
+	if (priv->font_hash){
+		g_hash_table_foreach_remove (priv->font_hash, kill_font, NULL);
+		g_hash_table_destroy (priv->font_hash);
+	}
+
+	for (i = 0; i < priv->pages->len; i++){
+		GnomeCanvasItem *page;
+		
+		page = g_array_index (priv->pages, GnomeCanvasItem *, i);
+		gtk_object_destroy (GTK_OBJECT (page));
+	}
+	
+	g_array_free (priv->pages, TRUE);
 
 	g_free (priv);
 	GTK_OBJECT_CLASS (print_preview_parent_class)->finalize (object);
